#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>

#include <glib.h>

#include <libxml/uri.h>

#include "debug.h"
#include "ftp.h"

gint    get_response          (gint sock_fd, GString *response,
                               Linelog *linelog);
gint    pasv_data_connection  (gint sock_fd, GString *buffer, Linelog *linelog);
gint    ftp_get_list          (gint sock_fd, gchar *path, GString *response,
                               GString *buffer,
                               Linelog *linelog);
gint    ftp_resolve_link      (char *filepath, GString *response,
                               GString *buffer);
void    ftp_parse_ls_response (GString *response, gulong *size,
                               Linelog *linelog);

/* returns 0 on success
 * returns 1 on error
 */
gint
ftp_login (gint sock_fd, xmlURIPtr url, Linelog *linelog, GError **err)
{
	GString *buffer = g_string_new(NULL);
    GString *logger;
    GError *error = NULL;
	int reply_code = 0;
	int login_error = 0;
    gchar** user_pass = NULL;

    get_response (sock_fd, buffer, linelog);

	if (atoi(buffer->str) == 220)
    {
        logger = g_string_new (NULL);
        g_string_printf (logger, "* 220 -> Service ready for new user\n");
        linelog_add (linelog, logger, &error);
    }

    if (url->user)
    {
        user_pass = g_strsplit (url->user, ":", 2);
    }

//FIXME: change to use name and password in the download object
    g_string_printf(buffer, "USER %s\r\n",
                    (user_pass && user_pass[0]) ? user_pass[0] : "anonymous");

	if (write (sock_fd, buffer->str, buffer->len) == -1)
	{
		//FIXME: log the error
		perror("Can't log in. USER failed");
		login_error = 1;
    }
	else
	{
        logger = g_string_new (NULL);
        g_string_printf (logger, "S %s", buffer->str);
        linelog_add (linelog, logger, &error);

        get_response(sock_fd, buffer, linelog);

		reply_code = atoi(buffer->str);
	}

    /* Valid respones after an USER command
       230 -> User logged in, proceed.
       331 -> User name okay, need password.
       332 -> Need account for login.
       421 -> Service not available, closing control connection.
       500 -> Syntax error, command unrecognized.
       501 -> Syntax error in parameters or arguments.
       530 -> Not logged in.
     */

    switch (reply_code)
    {
        case 230:
            break;
        case 331:
            break;
        case 332:
        case 421:
        case 500:
        case 501:
        case 530:
            login_error = 1;
            g_set_error (err, DM_DOWNMAND_ERROR, 1, buffer->str + 4);
            break;
        default:
    }

	if ((!login_error) && (reply_code == 331))
	{
        logger = g_string_new (NULL);
        g_string_printf (logger, "* 331 -> User name okay, need password\n");
        linelog_add (linelog, logger, &error);

//FIXME: change to use name and password in the download object
		g_string_printf(buffer, "PASS %s\r\n",
                        (user_pass && user_pass[1]) ? user_pass[1]
                        : "downman@downman.org");

		if (write(sock_fd, buffer->str, buffer->len) == -1)
		{
			//FIXME: log the error
			perror("Can't log in. PASS failed");
			login_error = 2;
		}
		else
		{
            logger = g_string_new (NULL);
            g_string_printf (logger, "S %s", buffer->str);
            linelog_add (linelog, logger, &error);

			get_response(sock_fd, buffer, linelog);

			reply_code = atoi(buffer->str);

			if (reply_code == 230)
			{
                logger = g_string_new (NULL);
                g_string_printf (logger, "* 230 -> User logged in, proceed\n");
                linelog_add (linelog, logger, &error);
			}
			else
			{
				login_error = 3;
			}
		}
	}

    if (user_pass)
        g_strfreev (user_pass);

    g_string_printf (buffer, "TYPE I\r\n");

    if (write (sock_fd, buffer->str, buffer->len) == -1)
    {
        //FIXME: log the error
        perror ("TYPE I failed");
        login_error = 4;
    }
    else
    {
        logger = g_string_new (NULL);
        g_string_printf (logger, "S %s", buffer->str);
        linelog_add (linelog, logger, &error);

        get_response (sock_fd, buffer, linelog);

        reply_code = atoi (buffer->str);
    }

    g_string_free (buffer, TRUE);
    return login_error;
}

gint
ftp_get_file_size (gint sock_fd, xmlURIPtr url, gulong *size, Linelog *linelog)
{
	GString *buffer = g_string_new(NULL);
	GString *response = g_string_new(NULL);
    GString *logger;
    GError *error = NULL;
	int reply_code = 0;
	int get_size_error = 0;

	g_string_printf(buffer, "TYPE I\r\n");

	if (write(sock_fd, buffer->str, buffer->len) == -1)
	{
		//FIXME: log the error
		perror("TYPE I failed");
		get_size_error = 1;
	}
	else
	{
        logger = g_string_new (NULL);
        g_string_printf (logger, "S %s", buffer->str);
        linelog_add (linelog, logger, &error);

		get_response (sock_fd, buffer, linelog);

		reply_code = atoi(buffer->str);
	}

	if ((!get_size_error) && (reply_code == 200))
	{

        logger = g_string_new (NULL);
        g_string_printf (logger, "* 200 Command okay\n");
        linelog_add (linelog, logger, &error);

		ftp_get_list (sock_fd, url->path, response, buffer, linelog);

        /* if the file is a link */
        if (response->str[0] == 'l')
        {
            char *link;
            ftp_resolve_link (url->path, response, buffer);
            link = strdup (buffer->str);
            ftp_get_list (sock_fd, link, response, buffer, linelog);
            free (link);
        }

        ftp_parse_ls_response (response, size, linelog);
	}

	g_string_free(response, TRUE);
	g_string_free(buffer, TRUE);
	return get_size_error;
}

gint
ftp_get (gint sock_fd, xmlURIPtr url, gint *data_fd, gulong file_offset,
         Linelog *linelog)
{
	GString *buffer = g_string_new(NULL);
    GString *logger;
    GError *error = NULL;
	int reply_code = 0;

	*data_fd = pasv_data_connection (sock_fd, buffer, linelog);

	g_string_printf(buffer, "REST %li\r\n", file_offset);

	if (write(sock_fd, buffer->str, buffer->len) == -1)
	{
		//FIXME: log the error
		perror("REST failed");
		return 49;
	}

    logger = g_string_new (NULL);
    g_string_printf (logger, "S %s", buffer->str);
    linelog_add (linelog, logger, &error);

    get_response (sock_fd, buffer, linelog);

	reply_code = atoi(buffer->str);
		
	g_string_printf(buffer, "RETR %s\r\n", url->path);
	
	if (write(sock_fd, buffer->str, buffer->len) == -1)
	{
		//FIXME: log the error
		perror("RETR failed");
		return 4;
	}

    logger = g_string_new (NULL);
    g_string_printf (logger, "S %s", buffer->str);
    linelog_add (linelog, logger, &error);

	get_response (sock_fd, buffer, linelog);

	reply_code = atoi(buffer->str);
	
	if (reply_code == 125)
	{
        logger = g_string_new (NULL);
        g_string_printf (logger, "* 125 -> Data connection already open; transfer starting\n");
        linelog_add (linelog, logger, &error);
	}

	g_string_free(buffer, TRUE);
	return 0;
}

/* 
*/
gint
get_response (gint sock_fd, GString *response, Linelog *linelog)
{
	GString *buffer = g_string_new(NULL);
    GString *logger;
    GError *error = NULL;
	GIOChannel* channel = g_io_channel_unix_new(sock_fd);
    guint code, command_code = 0;
    gboolean command_end = FALSE;

	g_io_channel_set_line_term(channel, "\r\n", -1);
	g_io_channel_set_encoding(channel, NULL, NULL);
	g_string_truncate (response, 0);

	/* Read single or multiline responses */
	do {
        code = 0;
        g_io_channel_read_line_string (channel, buffer, NULL, NULL);
        if (isdigit (buffer->str[0]))
        {
            if (!sscanf(buffer->str, "%u", &code))
            code = 0;
        }
        else
            code = 0;
    if (code)
    {
        if (command_code)
        {
            if (command_code != code)
            {
                command_end = TRUE;
            }
            else if (buffer->str[3] != '-')
            {
            	command_end = TRUE;
            }
        }
        else
        {
            command_code = code;
            if (buffer->str[3] != '-')
            {
                command_end = TRUE;
            }
        }
    }
		response = g_string_append (response, buffer->str);

        logger = g_string_new (NULL);
        g_string_printf (logger, "R %s%s", buffer->str,
                         (buffer->str[buffer->len - 1] == '\n')? "" : "\n");
        linelog_add (linelog, logger, &error);
	} while (!command_end);

	return 0;
}

/* returns the fd of the data connection
 */
gint
pasv_data_connection (gint sock_fd, GString *buffer, Linelog *linelog)
{
    GString *logger;
    GError *error = NULL;
	int nums[6];
	int reply_code = 0;
	int data_fd = -1;
	struct sockaddr_in addr;

	g_string_printf(buffer, "PASV\r\n");

	if (write(sock_fd, buffer->str, buffer->len) == -1)
	{
		//FIXME: log the error
		perror("PASV failed");
		return -1;
	}

    logger = g_string_new (NULL);
    g_string_printf (logger, "S %s", buffer->str);
    linelog_add (linelog, logger, &error);

	get_response (sock_fd, buffer, linelog);

	reply_code = atoi(buffer->str);

	if (reply_code == 227)
	{
		char *buf_ptr = buffer->str + 4;
		while (*buf_ptr && !isdigit(*buf_ptr))
			buf_ptr++;

        logger = g_string_new (NULL);
        g_string_printf (logger, "* 227 -> Entering Passive Mode (h1,h2,h3,h4,p1,p2)\n");
        linelog_add (linelog, logger, &error);

		if (sscanf(buf_ptr, "%d,%d,%d,%d,%d,%d", &nums[0], &nums[1], &nums[2],
			&nums[3], &nums[4], &nums[5]) != 6)
		{
			//FIXME: log the error
			perror("wrong PASV response");
			return 4;
		}

		addr.sin_family = AF_INET;
		addr.sin_addr.s_addr = htonl ((nums[0] << 24) | (nums[1] << 16) | (nums[2] << 8) | nums[3]);
		addr.sin_port = htons ((nums[4] << 8) | nums[5]);

		data_fd = socket(PF_INET, SOCK_STREAM, 0);
		if (data_fd < 0)
		{
			g_debug("DATA SOCKET: Can't create socket!\n");
            return -1;
		}

		connect(data_fd, (struct sockaddr *) &addr, sizeof(addr));
	}

	return data_fd;
}

gint
ftp_get_list (gint sock_fd, gchar *path, GString *response, GString *buffer,
              Linelog *linelog)
{
    GString *logger;
    GError *error = NULL;
	GIOChannel* channel;
	int data_fd;
	int get_list_error = 0;
	int reply_code = 0;

	data_fd = pasv_data_connection (sock_fd, buffer, linelog);

	channel = g_io_channel_unix_new (data_fd);
	g_io_channel_set_line_term (channel, "\r\n", -1);
	g_io_channel_set_encoding (channel, NULL, NULL);

	g_string_printf(buffer, "LIST %s\r\n", path);

	if (write(sock_fd, buffer->str, buffer->len) == -1)
	{
		//FIXME: log the error
		perror("LIST failed");
		get_list_error = 1;
	}
	else
	{
        logger = g_string_new (NULL);
        g_string_printf (logger, "S %s", buffer->str);
        linelog_add (linelog, logger, &error);

		get_response (sock_fd, buffer, linelog);

		reply_code = atoi (buffer->str);

		if ((reply_code == 125) || (reply_code == 150))
		{
			switch (reply_code)
			{
				case 125:
                    logger = g_string_new (NULL);
                    g_string_printf (logger, "* 125 -> Data connection already open; transfer starting\n");
                    linelog_add (linelog, logger, &error);
					break;
				case 150:
                    logger = g_string_new (NULL);
                    g_string_printf (logger, "* 150 -> File status okay; about to open data connection\n");
                    linelog_add (linelog, logger, &error);
					break;
			}
			g_string_truncate (response, 0);
			g_io_channel_read_line_string (channel, response, NULL, NULL);
			g_io_channel_close (channel);

            logger = g_string_new (NULL);
            g_string_printf (logger, "R %s\n", response->str);
            linelog_add (linelog, logger, &error);

			get_response (sock_fd, buffer, linelog);

			reply_code = atoi(buffer->str);

			if (reply_code != 226)
			{
				get_list_error = 2;
			}
		}
		else
		{
			get_list_error = 3;
		}
	}

    return get_list_error;
}

int ftp_resolve_link(char *filepath, GString *response, GString *buffer)
{
	char *link;

	g_string_truncate(buffer, 0);
	link = strstr(response->str, " -> ");
	if (link)
	{
		link += 4;
		if (link[0] == '/')
		{
			g_string_append(buffer, link);
		}
		else
		{
			char *file;
			g_string_append(buffer, filepath);
			file = strrchr(buffer->str, '/');
			if (file) file[1] = '\0';
			file = strrchr(response->str, '\n');
			if (file) file[0] = '\0';
			file = strrchr(response->str, '\r');
			if (file) file[0] = '\0';
			g_string_truncate(buffer, strlen(buffer->str));
			file = xmlBuildURI (link, buffer->str);

			g_string_printf(buffer, "%s", file);
		}
	}

//printf("NEWPATH = %s\n", buffer->str);
	return 0;
}

void
ftp_parse_ls_response (GString *response, gulong *size, Linelog *linelog)
{
    GString *logger;
    GError *error = NULL;
	int size_ok = 0;

	/* ftp.microsoft.com */
	if (sscanf(response->str, "%*s %*i %*s %*s %li", size) == 1)
	{
		size_ok = 1;
	}
	/* ftp.gnome.org */
	else if (sscanf(response->str, "%*s %*i %*s %li", size) == 1)
	{
		size_ok = 1;
	}

    logger = g_string_new (NULL);
    g_string_printf (logger, "* Size of file is %li\n", *size);
    linelog_add (linelog, logger, &error);
}
